@@ -33,22 +33,23 @@
 #![forbid(unsafe_code)]
 #![deny(missing_docs)]
 
+use dedupe::DedupeContext;
 use git2::{Config, Cred, CredentialHelper, RemoteCallbacks};
 use semver::Version as SemverVersion;
 use serde_derive::{Deserialize, Serialize};
 use smartstring::alias::String as SmolStr;
 use std::collections::HashMap;
-use std::collections::HashSet;
 use std::io;
 use std::path::Path;
 use std::sync::Arc;
 
 mod bare_index;
 mod error;
+mod dedupe;
 
 pub use bare_index::Crates;
 pub use bare_index::Index;
-
+pub use error::CratesIterError;
 pub use error::Error;
 
 static INDEX_GIT_URL: &str = "https://github.com/rust-lang/crates.io-index";
@@ -275,23 +276,22 @@ fn crate_prefix(accumulator: &mut String, crate_name: &str, separator: char) ->
         3 => {
             accumulator.push('3');
             accumulator.push(separator);
-            accumulator.push_str(crate_name.get(0..1)?);
+            accumulator.extend(crate_name.as_bytes().get(0..1)?.iter().map(|c| c.to_ascii_lowercase() as char));
         }
         _ => {
-            accumulator.push_str(crate_name.get(0..2)?);
+            accumulator.extend(crate_name.as_bytes().get(0..2)?.iter().map(|c| c.to_ascii_lowercase() as char));
             accumulator.push(separator);
-            accumulator.push_str(crate_name.get(2..4)?);
+            accumulator.extend(crate_name.as_bytes().get(2..4)?.iter().map(|c| c.to_ascii_lowercase() as char));
         }
     };
     Some(())
 }
 
 fn crate_name_to_relative_path(crate_name: &str) -> Option<String> {
-    let name_lower = crate_name.to_ascii_lowercase();
     let mut rel_path = String::with_capacity(crate_name.len() + 6);
-    crate_prefix(&mut rel_path, &name_lower, std::path::MAIN_SEPARATOR)?;
+    crate_prefix(&mut rel_path, crate_name, std::path::MAIN_SEPARATOR)?;
     rel_path.push(std::path::MAIN_SEPARATOR);
-    rel_path.push_str(&name_lower);
+    rel_path.extend(crate_name.as_bytes().iter().map(|c| c.to_ascii_lowercase() as char));
 
     Some(rel_path)
 }
@@ -313,7 +313,14 @@ impl Crate {
     }
 
     /// Parse crate file from in-memory JSON data
-    pub fn from_slice(mut bytes: &[u8]) -> io::Result<Crate> {
+    #[inline]
+    pub fn from_slice(bytes: &[u8]) -> io::Result<Crate> {
+        let mut dedupe = DedupeContext::new();
+        Self::from_slice_with_context(bytes, &mut dedupe)
+    }
+
+    /// Parse crate file from in-memory JSON data
+    pub(crate) fn from_slice_with_context(mut bytes: &[u8], dedupe: &mut DedupeContext) -> io::Result<Crate> {
         // Trim last newline
         while bytes.last() == Some(&b'\n') {
             bytes = &bytes[..bytes.len() - 1];
@@ -324,8 +331,6 @@ impl Crate {
             c == b'\n'
         }
         let num_versions = bytes.split(is_newline).count();
-        let mut deps_dedupe = HashSet::with_capacity(num_versions);
-        let mut features_dedupe = Vec::new();
         let mut versions = Vec::with_capacity(num_versions);
         for line in bytes.split(is_newline) {
             let mut version: Version = serde_json::from_slice(line)
@@ -340,16 +345,9 @@ impl Crate {
             }
 
             // Many versions have identical dependencies and features
-            if let Some(has_deps) = deps_dedupe.get(&version.deps) {
-                version.deps = Arc::clone(has_deps);
-            } else {
-                deps_dedupe.insert(Arc::clone(&version.deps));
-            }
-            if let Some(has_feats) = features_dedupe.iter().find(|v| *v == &version.features) {
-                version.features = Arc::clone(has_feats);
-            } else {
-                features_dedupe.push(Arc::clone(&version.features));
-            }
+            dedupe.deps(&mut version.deps);
+            dedupe.features(&mut version.features);
+
             versions.push(version);
         }
         if versions.is_empty() {
@@ -622,9 +620,10 @@ mod test {
         let mut found_gcc_crate = false;
 
         let index = Index::with_path(tmp_dir.path(), crate::INDEX_GIT_URL).unwrap();
+        let mut ctx = DedupeContext::new();
 
         for c in index.crates_refs().unwrap() {
-            match c.parse() {
+            match c.parse(&mut ctx) {
                 Ok(c) => {
                     if c.name() == "gcc" {
                         found_gcc_crate = true;
